# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/pubsub.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as src_dot_ray_dot_protobuf_dot_common__pb2
from . import dependency_pb2 as src_dot_ray_dot_protobuf_dot_dependency__pb2
from . import gcs_pb2 as src_dot_ray_dot_protobuf_dot_gcs__pb2
from . import logging_pb2 as src_dot_ray_dot_protobuf_dot_logging__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dsrc/ray/protobuf/pubsub.proto\x12\x07ray.rpc\x1a\x1dsrc/ray/protobuf/common.proto\x1a!src/ray/protobuf/dependency.proto\x1a\x1asrc/ray/protobuf/gcs.proto\x1a\x1esrc/ray/protobuf/logging.proto\"\xac\t\n\nPubMessage\x12\x37\n\x0c\x63hannel_type\x18\x01 \x01(\x0e\x32\x14.ray.rpc.ChannelTypeR\x0b\x63hannelType\x12\x15\n\x06key_id\x18\x02 \x01(\x0cR\x05keyId\x12k\n\x1eworker_object_eviction_message\x18\x03 \x01(\x0b\x32$.ray.rpc.WorkerObjectEvictionMessageH\x00R\x1bworkerObjectEvictionMessage\x12_\n\x1aworker_ref_removed_message\x18\x04 \x01(\x0b\x32 .ray.rpc.WorkerRefRemovedMessageH\x00R\x17workerRefRemovedMessage\x12q\n\x1fworker_object_locations_message\x18\x05 \x01(\x0b\x32(.ray.rpc.WorkerObjectLocationsPubMessageH\x00R\x1cworkerObjectLocationsMessage\x12>\n\ractor_message\x18\x07 \x01(\x0b\x32\x17.ray.rpc.ActorTableDataH\x00R\x0c\x61\x63torMessage\x12\x38\n\x0bjob_message\x18\x08 \x01(\x0b\x32\x15.ray.rpc.JobTableDataH\x00R\njobMessage\x12\x42\n\x11node_info_message\x18\t \x01(\x0b\x32\x14.ray.rpc.GcsNodeInfoH\x00R\x0fnodeInfoMessage\x12Q\n\x15node_resource_message\x18\n \x01(\x0b\x32\x1b.ray.rpc.NodeResourceChangeH\x00R\x13nodeResourceMessage\x12L\n\x14worker_delta_message\x18\x0b \x01(\x0b\x32\x18.ray.rpc.WorkerDeltaDataH\x00R\x12workerDeltaMessage\x12G\n\x12\x65rror_info_message\x18\x0c \x01(\x0b\x32\x17.ray.rpc.ErrorTableDataH\x00R\x10\x65rrorInfoMessage\x12?\n\x11log_batch_message\x18\r \x01(\x0b\x32\x11.ray.rpc.LogBatchH\x00R\x0flogBatchMessage\x12Q\n\x17python_function_message\x18\x0e \x01(\x0b\x32\x17.ray.rpc.PythonFunctionH\x00R\x15pythonFunctionMessage\x12[\n\x1bnode_resource_usage_message\x18\x0f \x01(\x0b\x32\x1a.ray.rpc.NodeResourceUsageH\x00R\x18nodeResourceUsageMessage\x12\x42\n\x0f\x66\x61ilure_message\x18\x06 \x01(\x0b\x32\x17.ray.rpc.FailureMessageH\x00R\x0e\x66\x61ilureMessage\x12\x1f\n\x0bsequence_id\x18\x10 \x01(\x03R\nsequenceIdB\x0f\n\rinner_message\":\n\x1bWorkerObjectEvictionMessage\x12\x1b\n\tobject_id\x18\x01 \x01(\x0cR\x08objectId\"]\n\x17WorkerRefRemovedMessage\x12\x42\n\rborrowed_refs\x18\x01 \x03(\x0b\x32\x1d.ray.rpc.ObjectReferenceCountR\x0c\x62orrowedRefs\"\x9a\x02\n\x1fWorkerObjectLocationsPubMessage\x12\x19\n\x08node_ids\x18\x01 \x03(\x0cR\x07nodeIds\x12\x1f\n\x0bobject_size\x18\x02 \x01(\x04R\nobjectSize\x12\x1f\n\x0bspilled_url\x18\x03 \x01(\tR\nspilledUrl\x12&\n\x0fspilled_node_id\x18\x04 \x01(\x0cR\rspilledNodeId\x12&\n\x0fprimary_node_id\x18\x06 \x01(\x0cR\rprimaryNodeId\x12\x1f\n\x0bref_removed\x18\x07 \x01(\x08R\nrefRemoved\x12)\n\x10pending_creation\x18\x08 \x01(\x08R\x0fpendingCreation\"\x10\n\x0e\x46\x61ilureMessage\"\x87\x02\n\x07\x43ommand\x12\x37\n\x0c\x63hannel_type\x18\x01 \x01(\x0e\x32\x14.ray.rpc.ChannelTypeR\x0b\x63hannelType\x12\x15\n\x06key_id\x18\x02 \x01(\x0cR\x05keyId\x12N\n\x13unsubscribe_message\x18\x03 \x01(\x0b\x32\x1b.ray.rpc.UnsubscribeMessageH\x00R\x12unsubscribeMessage\x12\x42\n\x11subscribe_message\x18\x04 \x01(\x0b\x32\x13.ray.rpc.SubMessageH\x00R\x10subscribeMessageB\x18\n\x16\x63ommand_message_one_of\"\x14\n\x12UnsubscribeMessage\"\xe9\x02\n\nSubMessage\x12n\n\x1eworker_object_eviction_message\x18\x01 \x01(\x0b\x32\'.ray.rpc.WorkerObjectEvictionSubMessageH\x00R\x1bworkerObjectEvictionMessage\x12\x62\n\x1aworker_ref_removed_message\x18\x02 \x01(\x0b\x32#.ray.rpc.WorkerRefRemovedSubMessageH\x00R\x17workerRefRemovedMessage\x12q\n\x1fworker_object_locations_message\x18\x03 \x01(\x0b\x32(.ray.rpc.WorkerObjectLocationsSubMessageH\x00R\x1cworkerObjectLocationsMessageB\x14\n\x12sub_message_one_of\"\xe5\x01\n\x1eWorkerObjectEvictionSubMessage\x12,\n\x12intended_worker_id\x18\x01 \x01(\x0cR\x10intendedWorkerId\x12\x1b\n\tobject_id\x18\x02 \x01(\x0cR\x08objectId\x12?\n\x12subscriber_address\x18\x03 \x01(\x0b\x32\x10.ray.rpc.AddressR\x11subscriberAddress\x12&\n\x0cgenerator_id\x18\x04 \x01(\x0cH\x00R\x0bgeneratorId\x88\x01\x01\x42\x0f\n\r_generator_id\"\xdc\x01\n\x1aWorkerRefRemovedSubMessage\x12,\n\x12intended_worker_id\x18\x01 \x01(\x0cR\x10intendedWorkerId\x12\x36\n\treference\x18\x02 \x01(\x0b\x32\x18.ray.rpc.ObjectReferenceR\treference\x12&\n\x0f\x63ontained_in_id\x18\x03 \x01(\x0cR\rcontainedInId\x12\x30\n\x14subscriber_worker_id\x18\x04 \x01(\x0cR\x12subscriberWorkerId\"l\n\x1fWorkerObjectLocationsSubMessage\x12,\n\x12intended_worker_id\x18\x01 \x01(\x0cR\x10intendedWorkerId\x12\x1b\n\tobject_id\x18\x02 \x01(\x0cR\x08objectId\"\x9d\x01\n\x18PubsubLongPollingRequest\x12#\n\rsubscriber_id\x18\x01 \x01(\x0cR\x0csubscriberId\x12\x39\n\x19max_processed_sequence_id\x18\x02 \x01(\x03R\x16maxProcessedSequenceId\x12!\n\x0cpublisher_id\x18\x03 \x01(\x0cR\x0bpublisherId\"s\n\x16PubsubLongPollingReply\x12\x36\n\x0cpub_messages\x18\x01 \x03(\x0b\x32\x13.ray.rpc.PubMessageR\x0bpubMessages\x12!\n\x0cpublisher_id\x18\x02 \x01(\x0cR\x0bpublisherId\"n\n\x19PubsubCommandBatchRequest\x12#\n\rsubscriber_id\x18\x01 \x01(\x0cR\x0csubscriberId\x12,\n\x08\x63ommands\x18\x02 \x03(\x0b\x32\x10.ray.rpc.CommandR\x08\x63ommands\"\x19\n\x17PubsubCommandBatchReply*\xca\x02\n\x0b\x43hannelType\x12\x1a\n\x16WORKER_OBJECT_EVICTION\x10\x00\x12\x1e\n\x1aWORKER_REF_REMOVED_CHANNEL\x10\x01\x12#\n\x1fWORKER_OBJECT_LOCATIONS_CHANNEL\x10\x02\x12\x15\n\x11GCS_ACTOR_CHANNEL\x10\x03\x12\x13\n\x0fGCS_JOB_CHANNEL\x10\x04\x12\x19\n\x15GCS_NODE_INFO_CHANNEL\x10\x05\x12\x1c\n\x18GCS_WORKER_DELTA_CHANNEL\x10\x06\x12\x1a\n\x16RAY_ERROR_INFO_CHANNEL\x10\x07\x12\x13\n\x0fRAY_LOG_CHANNEL\x10\x08\x12\x1f\n\x1bRAY_PYTHON_FUNCTION_CHANNEL\x10\t\x12#\n\x1fRAY_NODE_RESOURCE_USAGE_CHANNEL\x10\n2\xc8\x01\n\x11SubscriberService\x12W\n\x11PubsubLongPolling\x12!.ray.rpc.PubsubLongPollingRequest\x1a\x1f.ray.rpc.PubsubLongPollingReply\x12Z\n\x12PubsubCommandBatch\x12\".ray.rpc.PubsubCommandBatchRequest\x1a .ray.rpc.PubsubCommandBatchReplyB\x03\xf8\x01\x01\x62\x06proto3')

_CHANNELTYPE = DESCRIPTOR.enum_types_by_name['ChannelType']
ChannelType = enum_type_wrapper.EnumTypeWrapper(_CHANNELTYPE)
WORKER_OBJECT_EVICTION = 0
WORKER_REF_REMOVED_CHANNEL = 1
WORKER_OBJECT_LOCATIONS_CHANNEL = 2
GCS_ACTOR_CHANNEL = 3
GCS_JOB_CHANNEL = 4
GCS_NODE_INFO_CHANNEL = 5
GCS_WORKER_DELTA_CHANNEL = 6
RAY_ERROR_INFO_CHANNEL = 7
RAY_LOG_CHANNEL = 8
RAY_PYTHON_FUNCTION_CHANNEL = 9
RAY_NODE_RESOURCE_USAGE_CHANNEL = 10


_PUBMESSAGE = DESCRIPTOR.message_types_by_name['PubMessage']
_WORKEROBJECTEVICTIONMESSAGE = DESCRIPTOR.message_types_by_name['WorkerObjectEvictionMessage']
_WORKERREFREMOVEDMESSAGE = DESCRIPTOR.message_types_by_name['WorkerRefRemovedMessage']
_WORKEROBJECTLOCATIONSPUBMESSAGE = DESCRIPTOR.message_types_by_name['WorkerObjectLocationsPubMessage']
_FAILUREMESSAGE = DESCRIPTOR.message_types_by_name['FailureMessage']
_COMMAND = DESCRIPTOR.message_types_by_name['Command']
_UNSUBSCRIBEMESSAGE = DESCRIPTOR.message_types_by_name['UnsubscribeMessage']
_SUBMESSAGE = DESCRIPTOR.message_types_by_name['SubMessage']
_WORKEROBJECTEVICTIONSUBMESSAGE = DESCRIPTOR.message_types_by_name['WorkerObjectEvictionSubMessage']
_WORKERREFREMOVEDSUBMESSAGE = DESCRIPTOR.message_types_by_name['WorkerRefRemovedSubMessage']
_WORKEROBJECTLOCATIONSSUBMESSAGE = DESCRIPTOR.message_types_by_name['WorkerObjectLocationsSubMessage']
_PUBSUBLONGPOLLINGREQUEST = DESCRIPTOR.message_types_by_name['PubsubLongPollingRequest']
_PUBSUBLONGPOLLINGREPLY = DESCRIPTOR.message_types_by_name['PubsubLongPollingReply']
_PUBSUBCOMMANDBATCHREQUEST = DESCRIPTOR.message_types_by_name['PubsubCommandBatchRequest']
_PUBSUBCOMMANDBATCHREPLY = DESCRIPTOR.message_types_by_name['PubsubCommandBatchReply']
PubMessage = _reflection.GeneratedProtocolMessageType('PubMessage', (_message.Message,), {
  'DESCRIPTOR' : _PUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubMessage)
  })
_sym_db.RegisterMessage(PubMessage)

WorkerObjectEvictionMessage = _reflection.GeneratedProtocolMessageType('WorkerObjectEvictionMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKEROBJECTEVICTIONMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerObjectEvictionMessage)
  })
_sym_db.RegisterMessage(WorkerObjectEvictionMessage)

WorkerRefRemovedMessage = _reflection.GeneratedProtocolMessageType('WorkerRefRemovedMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKERREFREMOVEDMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerRefRemovedMessage)
  })
_sym_db.RegisterMessage(WorkerRefRemovedMessage)

WorkerObjectLocationsPubMessage = _reflection.GeneratedProtocolMessageType('WorkerObjectLocationsPubMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKEROBJECTLOCATIONSPUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerObjectLocationsPubMessage)
  })
_sym_db.RegisterMessage(WorkerObjectLocationsPubMessage)

FailureMessage = _reflection.GeneratedProtocolMessageType('FailureMessage', (_message.Message,), {
  'DESCRIPTOR' : _FAILUREMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.FailureMessage)
  })
_sym_db.RegisterMessage(FailureMessage)

Command = _reflection.GeneratedProtocolMessageType('Command', (_message.Message,), {
  'DESCRIPTOR' : _COMMAND,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.Command)
  })
_sym_db.RegisterMessage(Command)

UnsubscribeMessage = _reflection.GeneratedProtocolMessageType('UnsubscribeMessage', (_message.Message,), {
  'DESCRIPTOR' : _UNSUBSCRIBEMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.UnsubscribeMessage)
  })
_sym_db.RegisterMessage(UnsubscribeMessage)

SubMessage = _reflection.GeneratedProtocolMessageType('SubMessage', (_message.Message,), {
  'DESCRIPTOR' : _SUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.SubMessage)
  })
_sym_db.RegisterMessage(SubMessage)

WorkerObjectEvictionSubMessage = _reflection.GeneratedProtocolMessageType('WorkerObjectEvictionSubMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKEROBJECTEVICTIONSUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerObjectEvictionSubMessage)
  })
_sym_db.RegisterMessage(WorkerObjectEvictionSubMessage)

WorkerRefRemovedSubMessage = _reflection.GeneratedProtocolMessageType('WorkerRefRemovedSubMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKERREFREMOVEDSUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerRefRemovedSubMessage)
  })
_sym_db.RegisterMessage(WorkerRefRemovedSubMessage)

WorkerObjectLocationsSubMessage = _reflection.GeneratedProtocolMessageType('WorkerObjectLocationsSubMessage', (_message.Message,), {
  'DESCRIPTOR' : _WORKEROBJECTLOCATIONSSUBMESSAGE,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerObjectLocationsSubMessage)
  })
_sym_db.RegisterMessage(WorkerObjectLocationsSubMessage)

PubsubLongPollingRequest = _reflection.GeneratedProtocolMessageType('PubsubLongPollingRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUBSUBLONGPOLLINGREQUEST,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubsubLongPollingRequest)
  })
_sym_db.RegisterMessage(PubsubLongPollingRequest)

PubsubLongPollingReply = _reflection.GeneratedProtocolMessageType('PubsubLongPollingReply', (_message.Message,), {
  'DESCRIPTOR' : _PUBSUBLONGPOLLINGREPLY,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubsubLongPollingReply)
  })
_sym_db.RegisterMessage(PubsubLongPollingReply)

PubsubCommandBatchRequest = _reflection.GeneratedProtocolMessageType('PubsubCommandBatchRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUBSUBCOMMANDBATCHREQUEST,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubsubCommandBatchRequest)
  })
_sym_db.RegisterMessage(PubsubCommandBatchRequest)

PubsubCommandBatchReply = _reflection.GeneratedProtocolMessageType('PubsubCommandBatchReply', (_message.Message,), {
  'DESCRIPTOR' : _PUBSUBCOMMANDBATCHREPLY,
  '__module__' : 'src.ray.protobuf.pubsub_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubsubCommandBatchReply)
  })
_sym_db.RegisterMessage(PubsubCommandBatchReply)

_SUBSCRIBERSERVICE = DESCRIPTOR.services_by_name['SubscriberService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _CHANNELTYPE._serialized_start=3459
  _CHANNELTYPE._serialized_end=3789
  _PUBMESSAGE._serialized_start=169
  _PUBMESSAGE._serialized_end=1365
  _WORKEROBJECTEVICTIONMESSAGE._serialized_start=1367
  _WORKEROBJECTEVICTIONMESSAGE._serialized_end=1425
  _WORKERREFREMOVEDMESSAGE._serialized_start=1427
  _WORKERREFREMOVEDMESSAGE._serialized_end=1520
  _WORKEROBJECTLOCATIONSPUBMESSAGE._serialized_start=1523
  _WORKEROBJECTLOCATIONSPUBMESSAGE._serialized_end=1805
  _FAILUREMESSAGE._serialized_start=1807
  _FAILUREMESSAGE._serialized_end=1823
  _COMMAND._serialized_start=1826
  _COMMAND._serialized_end=2089
  _UNSUBSCRIBEMESSAGE._serialized_start=2091
  _UNSUBSCRIBEMESSAGE._serialized_end=2111
  _SUBMESSAGE._serialized_start=2114
  _SUBMESSAGE._serialized_end=2475
  _WORKEROBJECTEVICTIONSUBMESSAGE._serialized_start=2478
  _WORKEROBJECTEVICTIONSUBMESSAGE._serialized_end=2707
  _WORKERREFREMOVEDSUBMESSAGE._serialized_start=2710
  _WORKERREFREMOVEDSUBMESSAGE._serialized_end=2930
  _WORKEROBJECTLOCATIONSSUBMESSAGE._serialized_start=2932
  _WORKEROBJECTLOCATIONSSUBMESSAGE._serialized_end=3040
  _PUBSUBLONGPOLLINGREQUEST._serialized_start=3043
  _PUBSUBLONGPOLLINGREQUEST._serialized_end=3200
  _PUBSUBLONGPOLLINGREPLY._serialized_start=3202
  _PUBSUBLONGPOLLINGREPLY._serialized_end=3317
  _PUBSUBCOMMANDBATCHREQUEST._serialized_start=3319
  _PUBSUBCOMMANDBATCHREQUEST._serialized_end=3429
  _PUBSUBCOMMANDBATCHREPLY._serialized_start=3431
  _PUBSUBCOMMANDBATCHREPLY._serialized_end=3456
  _SUBSCRIBERSERVICE._serialized_start=3792
  _SUBSCRIBERSERVICE._serialized_end=3992
# @@protoc_insertion_point(module_scope)
