deprecation_msg = (
    "`HuggingFaceTrainer`, `HuggingFacePredictor` and `HuggingFaceCheckpoint` "
    "have been renamed to "
    "`TransformersTrainer`, `TransformersPredictor` and `TransformersCheckpoint` "
    "respectively. Update your code to use the new import "
    "paths. This will raise an exception in the future."
)
