# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/experimental/instance_manager.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n4src/ray/protobuf/experimental/instance_manager.proto\x12\x12ray.rpc.autoscaler\"\xf6\x01\n\x0cInstanceType\x12\x1b\n\ttype_name\x18\x01 \x01(\tR\x08typeName\x12<\n\tnode_kind\x18\x02 \x01(\x0e\x32\x1f.ray.rpc.autoscaler.RayNodeKindR\x08nodeKind\x12M\n\tresources\x18\x03 \x03(\x0b\x32/.ray.rpc.autoscaler.InstanceType.ResourcesEntryR\tresources\x1a<\n\x0eResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"\"\n GetAvailableInstanceTypesRequest\"\x9d\x01\n!GetAvailableInstanceTypesResponse\x12G\n\x0einstance_types\x18\x01 \x03(\x0b\x32 .ray.rpc.autoscaler.InstanceTypeR\rinstanceTypes\x12/\n\x13\x61vailable_instances\x18\x02 \x01(\x03R\x12\x61vailableInstances\"\xca\x06\n\x08Instance\x12\x1f\n\x0binstance_id\x18\x0b \x01(\tR\ninstanceId\x12\x43\n\x06status\x18\x0c \x01(\x0e\x32+.ray.rpc.autoscaler.Instance.InstanceStatusR\x06status\x12#\n\rinstance_type\x18\r \x01(\tR\x0cinstanceType\x12Y\n\x0ftotal_resources\x18\x0e \x03(\x0b\x32\x30.ray.rpc.autoscaler.Instance.TotalResourcesEntryR\x0etotalResources\x12\x41\n\x1dtimestamp_since_last_modified\x18\x0f \x01(\x03R\x1atimestampSinceLastModified\x12*\n\x11\x63loud_instance_id\x18\x10 \x01(\tR\x0f\x63loudInstanceId\x12\x1f\n\x0binternal_ip\x18\x11 \x01(\tR\ninternalIp\x12\x1f\n\x0b\x65xternal_ip\x18\x12 \x01(\tR\nexternalIp\x12\x18\n\x07version\x18\x13 \x01(\x03R\x07version\x12\x45\n\nray_status\x18\x14 \x01(\x0e\x32&.ray.rpc.autoscaler.Instance.RayStatusR\trayStatus\x1a\x41\n\x13TotalResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"\x90\x01\n\x0eInstanceStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06QEUEUD\x10\x01\x12\r\n\tREQUESTED\x10\x02\x12\r\n\tALLOCATED\x10\x03\x12\x15\n\x11\x41LLOCATION_FAILED\x10\x04\x12\x0c\n\x08STOPPING\x10\x05\x12\x0b\n\x07STOPPED\x10\x06\x12\x15\n\x11GARBAGE_COLLECTED\x10\x07\"p\n\tRayStatus\x12\x15\n\x11RAY_STATUS_UNKOWN\x10\x00\x12\x12\n\x0eRAY_INSTALLING\x10\x01\x12\x16\n\x12RAY_INSTALL_FAILED\x10\x02\x12\x0f\n\x0bRAY_RUNNING\x10\x03\x12\x0f\n\x0bRAY_STOPPED\x10\x04\"\xd8\x01\n!UpdateInstanceManagerStateRequest\x12)\n\x10\x65xpected_version\x18\x01 \x01(\x03R\x0f\x65xpectedVersion\x12M\n\x12new_nodes_to_start\x18\x02 \x03(\x0b\x32 .ray.rpc.autoscaler.InstanceTypeR\x0fnewNodesToStart\x12\x39\n\x19instance_ids_to_terminate\x18\x03 \x03(\tR\x16instanceIdsToTerminate\"z\n\x1fUpdateInstanceManagerStateReply\x12\x18\n\x07success\x18\x01 \x01(\x08R\x07success\x12#\n\rerror_message\x18\x02 \x01(\tR\x0c\x65rrorMessage\x12\x18\n\x07version\x18\x03 \x01(\x03R\x07version\"l\n\x14InstanceManagerState\x12\x18\n\x07version\x18\x01 \x01(\x03R\x07version\x12:\n\tinstances\x18\x02 \x03(\x0b\x32\x1c.ray.rpc.autoscaler.InstanceR\tinstances\" \n\x1eGetInstanceManagerStateRequest\"^\n\x1cGetInstanceManagerStateReply\x12>\n\x05state\x18\x01 \x01(\x0b\x32(.ray.rpc.autoscaler.InstanceManagerStateR\x05state*0\n\x0bRayNodeKind\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x08\n\x04HEAD\x10\x01\x12\n\n\x06WORKER\x10\x02\x32\xaf\x03\n\x16InstanceManagerService\x12\x7f\n\x17GetInstanceManagerState\x12\x32.ray.rpc.autoscaler.GetInstanceManagerStateRequest\x1a\x30.ray.rpc.autoscaler.GetInstanceManagerStateReply\x12\x88\x01\n\x1aUpdateInstanceManagerState\x12\x35.ray.rpc.autoscaler.UpdateInstanceManagerStateRequest\x1a\x33.ray.rpc.autoscaler.UpdateInstanceManagerStateReply\x12\x88\x01\n\x19GetAvailableInstanceTypes\x12\x34.ray.rpc.autoscaler.GetAvailableInstanceTypesRequest\x1a\x35.ray.rpc.autoscaler.GetAvailableInstanceTypesResponseB\x03\xf8\x01\x01\x62\x06proto3')

_RAYNODEKIND = DESCRIPTOR.enum_types_by_name['RayNodeKind']
RayNodeKind = enum_type_wrapper.EnumTypeWrapper(_RAYNODEKIND)
UNKNOWN = 0
HEAD = 1
WORKER = 2


_INSTANCETYPE = DESCRIPTOR.message_types_by_name['InstanceType']
_INSTANCETYPE_RESOURCESENTRY = _INSTANCETYPE.nested_types_by_name['ResourcesEntry']
_GETAVAILABLEINSTANCETYPESREQUEST = DESCRIPTOR.message_types_by_name['GetAvailableInstanceTypesRequest']
_GETAVAILABLEINSTANCETYPESRESPONSE = DESCRIPTOR.message_types_by_name['GetAvailableInstanceTypesResponse']
_INSTANCE = DESCRIPTOR.message_types_by_name['Instance']
_INSTANCE_TOTALRESOURCESENTRY = _INSTANCE.nested_types_by_name['TotalResourcesEntry']
_UPDATEINSTANCEMANAGERSTATEREQUEST = DESCRIPTOR.message_types_by_name['UpdateInstanceManagerStateRequest']
_UPDATEINSTANCEMANAGERSTATEREPLY = DESCRIPTOR.message_types_by_name['UpdateInstanceManagerStateReply']
_INSTANCEMANAGERSTATE = DESCRIPTOR.message_types_by_name['InstanceManagerState']
_GETINSTANCEMANAGERSTATEREQUEST = DESCRIPTOR.message_types_by_name['GetInstanceManagerStateRequest']
_GETINSTANCEMANAGERSTATEREPLY = DESCRIPTOR.message_types_by_name['GetInstanceManagerStateReply']
_INSTANCE_INSTANCESTATUS = _INSTANCE.enum_types_by_name['InstanceStatus']
_INSTANCE_RAYSTATUS = _INSTANCE.enum_types_by_name['RayStatus']
InstanceType = _reflection.GeneratedProtocolMessageType('InstanceType', (_message.Message,), {

  'ResourcesEntry' : _reflection.GeneratedProtocolMessageType('ResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _INSTANCETYPE_RESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.experimental.instance_manager_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.InstanceType.ResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _INSTANCETYPE,
  '__module__' : 'src.ray.protobuf.experimental.instance_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.InstanceType)
  })
_sym_db.RegisterMessage(InstanceType)
_sym_db.RegisterMessage(InstanceType.ResourcesEntry)

GetAvailableInstanceTypesRequest = _reflection.GeneratedProtocolMessageType('GetAvailableInstanceTypesRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETAVAILABLEINSTANCETYPESREQUEST,
  '__module__' : 'src.ray.protobuf.experimental.instance_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.GetAvailableInstanceTypesRequest)
  })
_sym_db.RegisterMessage(GetAvailableInstanceTypesRequest)

GetAvailableInstanceTypesResponse = _reflection.GeneratedProtocolMessageType('GetAvailableInstanceTypesResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETAVAILABLEINSTANCETYPESRESPONSE,
  '__module__' : 'src.ray.protobuf.experimental.instance_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.GetAvailableInstanceTypesResponse)
  })
_sym_db.RegisterMessage(GetAvailableInstanceTypesResponse)

Instance = _reflection.GeneratedProtocolMessageType('Instance', (_message.Message,), {

  'TotalResourcesEntry' : _reflection.GeneratedProtocolMessageType('TotalResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _INSTANCE_TOTALRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.experimental.instance_manager_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.Instance.TotalResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _INSTANCE,
  '__module__' : 'src.ray.protobuf.experimental.instance_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.Instance)
  })
_sym_db.RegisterMessage(Instance)
_sym_db.RegisterMessage(Instance.TotalResourcesEntry)

UpdateInstanceManagerStateRequest = _reflection.GeneratedProtocolMessageType('UpdateInstanceManagerStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEINSTANCEMANAGERSTATEREQUEST,
  '__module__' : 'src.ray.protobuf.experimental.instance_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.UpdateInstanceManagerStateRequest)
  })
_sym_db.RegisterMessage(UpdateInstanceManagerStateRequest)

UpdateInstanceManagerStateReply = _reflection.GeneratedProtocolMessageType('UpdateInstanceManagerStateReply', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEINSTANCEMANAGERSTATEREPLY,
  '__module__' : 'src.ray.protobuf.experimental.instance_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.UpdateInstanceManagerStateReply)
  })
_sym_db.RegisterMessage(UpdateInstanceManagerStateReply)

InstanceManagerState = _reflection.GeneratedProtocolMessageType('InstanceManagerState', (_message.Message,), {
  'DESCRIPTOR' : _INSTANCEMANAGERSTATE,
  '__module__' : 'src.ray.protobuf.experimental.instance_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.InstanceManagerState)
  })
_sym_db.RegisterMessage(InstanceManagerState)

GetInstanceManagerStateRequest = _reflection.GeneratedProtocolMessageType('GetInstanceManagerStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETINSTANCEMANAGERSTATEREQUEST,
  '__module__' : 'src.ray.protobuf.experimental.instance_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.GetInstanceManagerStateRequest)
  })
_sym_db.RegisterMessage(GetInstanceManagerStateRequest)

GetInstanceManagerStateReply = _reflection.GeneratedProtocolMessageType('GetInstanceManagerStateReply', (_message.Message,), {
  'DESCRIPTOR' : _GETINSTANCEMANAGERSTATEREPLY,
  '__module__' : 'src.ray.protobuf.experimental.instance_manager_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.autoscaler.GetInstanceManagerStateReply)
  })
_sym_db.RegisterMessage(GetInstanceManagerStateReply)

_INSTANCEMANAGERSERVICE = DESCRIPTOR.services_by_name['InstanceManagerService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _INSTANCETYPE_RESOURCESENTRY._options = None
  _INSTANCETYPE_RESOURCESENTRY._serialized_options = b'8\001'
  _INSTANCE_TOTALRESOURCESENTRY._options = None
  _INSTANCE_TOTALRESOURCESENTRY._serialized_options = b'8\001'
  _RAYNODEKIND._serialized_start=1949
  _RAYNODEKIND._serialized_end=1997
  _INSTANCETYPE._serialized_start=77
  _INSTANCETYPE._serialized_end=323
  _INSTANCETYPE_RESOURCESENTRY._serialized_start=263
  _INSTANCETYPE_RESOURCESENTRY._serialized_end=323
  _GETAVAILABLEINSTANCETYPESREQUEST._serialized_start=325
  _GETAVAILABLEINSTANCETYPESREQUEST._serialized_end=359
  _GETAVAILABLEINSTANCETYPESRESPONSE._serialized_start=362
  _GETAVAILABLEINSTANCETYPESRESPONSE._serialized_end=519
  _INSTANCE._serialized_start=522
  _INSTANCE._serialized_end=1364
  _INSTANCE_TOTALRESOURCESENTRY._serialized_start=1038
  _INSTANCE_TOTALRESOURCESENTRY._serialized_end=1103
  _INSTANCE_INSTANCESTATUS._serialized_start=1106
  _INSTANCE_INSTANCESTATUS._serialized_end=1250
  _INSTANCE_RAYSTATUS._serialized_start=1252
  _INSTANCE_RAYSTATUS._serialized_end=1364
  _UPDATEINSTANCEMANAGERSTATEREQUEST._serialized_start=1367
  _UPDATEINSTANCEMANAGERSTATEREQUEST._serialized_end=1583
  _UPDATEINSTANCEMANAGERSTATEREPLY._serialized_start=1585
  _UPDATEINSTANCEMANAGERSTATEREPLY._serialized_end=1707
  _INSTANCEMANAGERSTATE._serialized_start=1709
  _INSTANCEMANAGERSTATE._serialized_end=1817
  _GETINSTANCEMANAGERSTATEREQUEST._serialized_start=1819
  _GETINSTANCEMANAGERSTATEREQUEST._serialized_end=1851
  _GETINSTANCEMANAGERSTATEREPLY._serialized_start=1853
  _GETINSTANCEMANAGERSTATEREPLY._serialized_end=1947
  _INSTANCEMANAGERSERVICE._serialized_start=2000
  _INSTANCEMANAGERSERVICE._serialized_end=2431
# @@protoc_insertion_point(module_scope)
