# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/serve.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1csrc/ray/protobuf/serve.proto\x12\tray.serve\"\xcd\x03\n\x11\x41utoscalingConfig\x12!\n\x0cmin_replicas\x18\x01 \x01(\rR\x0bminReplicas\x12!\n\x0cmax_replicas\x18\x02 \x01(\rR\x0bmaxReplicas\x12S\n\'target_num_ongoing_requests_per_replica\x18\x03 \x01(\x01R\"targetNumOngoingRequestsPerReplica\x12,\n\x12metrics_interval_s\x18\x04 \x01(\x01R\x10metricsIntervalS\x12+\n\x12look_back_period_s\x18\x05 \x01(\x01R\x0flookBackPeriodS\x12)\n\x10smoothing_factor\x18\x06 \x01(\x01R\x0fsmoothingFactor\x12*\n\x11\x64ownscale_delay_s\x18\x07 \x01(\x01R\x0f\x64ownscaleDelayS\x12&\n\x0fupscale_delay_s\x18\x08 \x01(\x01R\rupscaleDelayS\x12.\n\x10initial_replicas\x18\t \x01(\rH\x00R\x0finitialReplicas\x88\x01\x01\x42\x13\n\x11_initial_replicas\"\x99\x05\n\x10\x44\x65ploymentConfig\x12!\n\x0cnum_replicas\x18\x01 \x01(\x05R\x0bnumReplicas\x12\x34\n\x16max_concurrent_queries\x18\x02 \x01(\x05R\x14maxConcurrentQueries\x12\x1f\n\x0buser_config\x18\x03 \x01(\x0cR\nuserConfig\x12@\n\x1dgraceful_shutdown_wait_loop_s\x18\x04 \x01(\x01R\x19gracefulShutdownWaitLoopS\x12=\n\x1bgraceful_shutdown_timeout_s\x18\x05 \x01(\x01R\x18gracefulShutdownTimeoutS\x12\x31\n\x15health_check_period_s\x18\x06 \x01(\x01R\x12healthCheckPeriodS\x12\x33\n\x16health_check_timeout_s\x18\x07 \x01(\x01R\x13healthCheckTimeoutS\x12*\n\x11is_cross_language\x18\x08 \x01(\x08R\x0fisCrossLanguage\x12N\n\x13\x64\x65ployment_language\x18\t \x01(\x0e\x32\x1d.ray.serve.DeploymentLanguageR\x12\x64\x65ploymentLanguage\x12K\n\x12\x61utoscaling_config\x18\n \x01(\x0b\x32\x1c.ray.serve.AutoscalingConfigR\x11\x61utoscalingConfig\x12\x18\n\x07version\x18\x0b \x01(\tR\x07version\x12?\n\x1cuser_configured_option_names\x18\x0c \x03(\tR\x19userConfiguredOptionNames\"\xec\x01\n\x0fRequestMetadata\x12\x1d\n\nrequest_id\x18\x01 \x01(\tR\trequestId\x12\x1a\n\x08\x65ndpoint\x18\x02 \x01(\tR\x08\x65ndpoint\x12\x1f\n\x0b\x63\x61ll_method\x18\x03 \x01(\tR\ncallMethod\x12\x41\n\x07\x63ontext\x18\x04 \x03(\x0b\x32\'.ray.serve.RequestMetadata.ContextEntryR\x07\x63ontext\x1a:\n\x0c\x43ontextEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"$\n\x0eRequestWrapper\x12\x12\n\x04\x62ody\x18\x01 \x01(\x0cR\x04\x62ody\"Y\n\rUpdatedObject\x12\'\n\x0fobject_snapshot\x18\x01 \x01(\x0cR\x0eobjectSnapshot\x12\x1f\n\x0bsnapshot_id\x18\x02 \x01(\x05R\nsnapshotId\"\xbb\x01\n\x0fLongPollRequest\x12\x62\n\x14keys_to_snapshot_ids\x18\x01 \x03(\x0b\x32\x31.ray.serve.LongPollRequest.KeysToSnapshotIdsEntryR\x11keysToSnapshotIds\x1a\x44\n\x16KeysToSnapshotIdsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x05R\x05value:\x02\x38\x01\"\xc5\x01\n\x0eLongPollResult\x12V\n\x0fupdated_objects\x18\x01 \x03(\x0b\x32-.ray.serve.LongPollResult.UpdatedObjectsEntryR\x0eupdatedObjects\x1a[\n\x13UpdatedObjectsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12.\n\x05value\x18\x02 \x01(\x0b\x32\x18.ray.serve.UpdatedObjectR\x05value:\x02\x38\x01\"\xc1\x01\n\x0c\x45ndpointInfo\x12#\n\rendpoint_name\x18\x01 \x01(\tR\x0c\x65ndpointName\x12\x14\n\x05route\x18\x02 \x01(\tR\x05route\x12;\n\x06\x63onfig\x18\x03 \x03(\x0b\x32#.ray.serve.EndpointInfo.ConfigEntryR\x06\x63onfig\x1a\x39\n\x0b\x43onfigEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\xa9\x01\n\x0b\x45ndpointSet\x12\x43\n\tendpoints\x18\x01 \x03(\x0b\x32%.ray.serve.EndpointSet.EndpointsEntryR\tendpoints\x1aU\n\x0e\x45ndpointsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12-\n\x05value\x18\x02 \x01(\x0b\x32\x17.ray.serve.EndpointInfoR\x05value:\x02\x38\x01\"%\n\rActorNameList\x12\x14\n\x05names\x18\x01 \x03(\tR\x05names\"\xac\x01\n\x11\x44\x65ploymentVersion\x12!\n\x0c\x63ode_version\x18\x01 \x01(\tR\x0b\x63odeVersion\x12H\n\x11\x64\x65ployment_config\x18\x02 \x01(\x0b\x32\x1b.ray.serve.DeploymentConfigR\x10\x64\x65ploymentConfig\x12*\n\x11ray_actor_options\x18\x03 \x01(\tR\x0frayActorOptions\"\xd0\x01\n\rReplicaConfig\x12.\n\x13\x64\x65ployment_def_name\x18\x01 \x01(\tR\x11\x64\x65ploymentDefName\x12%\n\x0e\x64\x65ployment_def\x18\x02 \x01(\x0cR\rdeploymentDef\x12\x1b\n\tinit_args\x18\x03 \x01(\x0cR\x08initArgs\x12\x1f\n\x0binit_kwargs\x18\x04 \x01(\x0cR\ninitKwargs\x12*\n\x11ray_actor_options\x18\x05 \x01(\tR\x0frayActorOptions\"\xc7\x02\n\x0e\x44\x65ploymentInfo\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12H\n\x11\x64\x65ployment_config\x18\x02 \x01(\x0b\x32\x1b.ray.serve.DeploymentConfigR\x10\x64\x65ploymentConfig\x12?\n\x0ereplica_config\x18\x03 \x01(\x0b\x32\x18.ray.serve.ReplicaConfigR\rreplicaConfig\x12\"\n\rstart_time_ms\x18\x04 \x01(\x03R\x0bstartTimeMs\x12\x1d\n\nactor_name\x18\x05 \x01(\tR\tactorName\x12\x18\n\x07version\x18\x06 \x01(\tR\x07version\x12\x1e\n\x0b\x65nd_time_ms\x18\x07 \x01(\x03R\tendTimeMs\x12\x19\n\x08\x61pp_name\x18\x08 \x01(\tR\x07\x61ppName\"k\n\x0f\x44\x65ploymentRoute\x12\x42\n\x0f\x64\x65ployment_info\x18\x01 \x01(\x0b\x32\x19.ray.serve.DeploymentInfoR\x0e\x64\x65ploymentInfo\x12\x14\n\x05route\x18\x02 \x01(\tR\x05route\"^\n\x13\x44\x65ploymentRouteList\x12G\n\x11\x64\x65ployment_routes\x18\x01 \x03(\x0b\x32\x1a.ray.serve.DeploymentRouteR\x10\x64\x65ploymentRoutes\"y\n\x14\x44\x65ploymentStatusInfo\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x33\n\x06status\x18\x02 \x01(\x0e\x32\x1b.ray.serve.DeploymentStatusR\x06status\x12\x18\n\x07message\x18\x03 \x01(\tR\x07message\"s\n\x18\x44\x65ploymentStatusInfoList\x12W\n\x17\x64\x65ployment_status_infos\x18\x01 \x03(\x0b\x32\x1f.ray.serve.DeploymentStatusInfoR\x15\x64\x65ploymentStatusInfos\"\x9a\x01\n\x15\x41pplicationStatusInfo\x12\x34\n\x06status\x18\x01 \x01(\x0e\x32\x1c.ray.serve.ApplicationStatusR\x06status\x12\x18\n\x07message\x18\x02 \x01(\tR\x07message\x12\x31\n\x14\x64\x65ployment_timestamp\x18\x03 \x01(\x01R\x13\x64\x65ploymentTimestamp\"\xbb\x01\n\x0eStatusOverview\x12?\n\napp_status\x18\x01 \x01(\x0b\x32 .ray.serve.ApplicationStatusInfoR\tappStatus\x12T\n\x13\x64\x65ployment_statuses\x18\x02 \x01(\x0b\x32#.ray.serve.DeploymentStatusInfoListR\x12\x64\x65ploymentStatuses\x12\x12\n\x04name\x18\x03 \x01(\tR\x04name\"\x86\x01\n\x0ePredictRequest\x12:\n\x05input\x18\x02 \x03(\x0b\x32$.ray.serve.PredictRequest.InputEntryR\x05input\x1a\x38\n\nInputEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x0cR\x05value:\x02\x38\x01\"1\n\x0fPredictResponse\x12\x1e\n\nprediction\x18\x01 \x01(\x0cR\nprediction**\n\x12\x44\x65ploymentLanguage\x12\n\n\x06PYTHON\x10\x00\x12\x08\n\x04JAVA\x10\x01*r\n\x10\x44\x65ploymentStatus\x12\x1e\n\x1a\x44\x45PLOYMENT_STATUS_UPDATING\x10\x00\x12\x1d\n\x19\x44\x45PLOYMENT_STATUS_HEALTHY\x10\x01\x12\x1f\n\x1b\x44\x45PLOYMENT_STATUS_UNHEALTHY\x10\x02*\xe2\x01\n\x11\x41pplicationStatus\x12 \n\x1c\x41PPLICATION_STATUS_DEPLOYING\x10\x00\x12\x1e\n\x1a\x41PPLICATION_STATUS_RUNNING\x10\x01\x12$\n APPLICATION_STATUS_DEPLOY_FAILED\x10\x02\x12\x1f\n\x1b\x41PPLICATION_STATUS_DELETING\x10\x03\x12\"\n\x1e\x41PPLICATION_STATUS_NOT_STARTED\x10\x05\x12 \n\x1c\x41PPLICATION_STATUS_UNHEALTHY\x10\x06\x32V\n\x12PredictAPIsService\x12@\n\x07Predict\x12\x19.ray.serve.PredictRequest\x1a\x1a.ray.serve.PredictResponseB*\n\x16io.ray.serve.generatedB\x0bServeProtosP\x01\xf8\x01\x01\x62\x06proto3')

_DEPLOYMENTLANGUAGE = DESCRIPTOR.enum_types_by_name['DeploymentLanguage']
DeploymentLanguage = enum_type_wrapper.EnumTypeWrapper(_DEPLOYMENTLANGUAGE)
_DEPLOYMENTSTATUS = DESCRIPTOR.enum_types_by_name['DeploymentStatus']
DeploymentStatus = enum_type_wrapper.EnumTypeWrapper(_DEPLOYMENTSTATUS)
_APPLICATIONSTATUS = DESCRIPTOR.enum_types_by_name['ApplicationStatus']
ApplicationStatus = enum_type_wrapper.EnumTypeWrapper(_APPLICATIONSTATUS)
PYTHON = 0
JAVA = 1
DEPLOYMENT_STATUS_UPDATING = 0
DEPLOYMENT_STATUS_HEALTHY = 1
DEPLOYMENT_STATUS_UNHEALTHY = 2
APPLICATION_STATUS_DEPLOYING = 0
APPLICATION_STATUS_RUNNING = 1
APPLICATION_STATUS_DEPLOY_FAILED = 2
APPLICATION_STATUS_DELETING = 3
APPLICATION_STATUS_NOT_STARTED = 5
APPLICATION_STATUS_UNHEALTHY = 6


_AUTOSCALINGCONFIG = DESCRIPTOR.message_types_by_name['AutoscalingConfig']
_DEPLOYMENTCONFIG = DESCRIPTOR.message_types_by_name['DeploymentConfig']
_REQUESTMETADATA = DESCRIPTOR.message_types_by_name['RequestMetadata']
_REQUESTMETADATA_CONTEXTENTRY = _REQUESTMETADATA.nested_types_by_name['ContextEntry']
_REQUESTWRAPPER = DESCRIPTOR.message_types_by_name['RequestWrapper']
_UPDATEDOBJECT = DESCRIPTOR.message_types_by_name['UpdatedObject']
_LONGPOLLREQUEST = DESCRIPTOR.message_types_by_name['LongPollRequest']
_LONGPOLLREQUEST_KEYSTOSNAPSHOTIDSENTRY = _LONGPOLLREQUEST.nested_types_by_name['KeysToSnapshotIdsEntry']
_LONGPOLLRESULT = DESCRIPTOR.message_types_by_name['LongPollResult']
_LONGPOLLRESULT_UPDATEDOBJECTSENTRY = _LONGPOLLRESULT.nested_types_by_name['UpdatedObjectsEntry']
_ENDPOINTINFO = DESCRIPTOR.message_types_by_name['EndpointInfo']
_ENDPOINTINFO_CONFIGENTRY = _ENDPOINTINFO.nested_types_by_name['ConfigEntry']
_ENDPOINTSET = DESCRIPTOR.message_types_by_name['EndpointSet']
_ENDPOINTSET_ENDPOINTSENTRY = _ENDPOINTSET.nested_types_by_name['EndpointsEntry']
_ACTORNAMELIST = DESCRIPTOR.message_types_by_name['ActorNameList']
_DEPLOYMENTVERSION = DESCRIPTOR.message_types_by_name['DeploymentVersion']
_REPLICACONFIG = DESCRIPTOR.message_types_by_name['ReplicaConfig']
_DEPLOYMENTINFO = DESCRIPTOR.message_types_by_name['DeploymentInfo']
_DEPLOYMENTROUTE = DESCRIPTOR.message_types_by_name['DeploymentRoute']
_DEPLOYMENTROUTELIST = DESCRIPTOR.message_types_by_name['DeploymentRouteList']
_DEPLOYMENTSTATUSINFO = DESCRIPTOR.message_types_by_name['DeploymentStatusInfo']
_DEPLOYMENTSTATUSINFOLIST = DESCRIPTOR.message_types_by_name['DeploymentStatusInfoList']
_APPLICATIONSTATUSINFO = DESCRIPTOR.message_types_by_name['ApplicationStatusInfo']
_STATUSOVERVIEW = DESCRIPTOR.message_types_by_name['StatusOverview']
_PREDICTREQUEST = DESCRIPTOR.message_types_by_name['PredictRequest']
_PREDICTREQUEST_INPUTENTRY = _PREDICTREQUEST.nested_types_by_name['InputEntry']
_PREDICTRESPONSE = DESCRIPTOR.message_types_by_name['PredictResponse']
AutoscalingConfig = _reflection.GeneratedProtocolMessageType('AutoscalingConfig', (_message.Message,), {
  'DESCRIPTOR' : _AUTOSCALINGCONFIG,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.AutoscalingConfig)
  })
_sym_db.RegisterMessage(AutoscalingConfig)

DeploymentConfig = _reflection.GeneratedProtocolMessageType('DeploymentConfig', (_message.Message,), {
  'DESCRIPTOR' : _DEPLOYMENTCONFIG,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.DeploymentConfig)
  })
_sym_db.RegisterMessage(DeploymentConfig)

RequestMetadata = _reflection.GeneratedProtocolMessageType('RequestMetadata', (_message.Message,), {

  'ContextEntry' : _reflection.GeneratedProtocolMessageType('ContextEntry', (_message.Message,), {
    'DESCRIPTOR' : _REQUESTMETADATA_CONTEXTENTRY,
    '__module__' : 'src.ray.protobuf.serve_pb2'
    # @@protoc_insertion_point(class_scope:ray.serve.RequestMetadata.ContextEntry)
    })
  ,
  'DESCRIPTOR' : _REQUESTMETADATA,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.RequestMetadata)
  })
_sym_db.RegisterMessage(RequestMetadata)
_sym_db.RegisterMessage(RequestMetadata.ContextEntry)

RequestWrapper = _reflection.GeneratedProtocolMessageType('RequestWrapper', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTWRAPPER,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.RequestWrapper)
  })
_sym_db.RegisterMessage(RequestWrapper)

UpdatedObject = _reflection.GeneratedProtocolMessageType('UpdatedObject', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEDOBJECT,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.UpdatedObject)
  })
_sym_db.RegisterMessage(UpdatedObject)

LongPollRequest = _reflection.GeneratedProtocolMessageType('LongPollRequest', (_message.Message,), {

  'KeysToSnapshotIdsEntry' : _reflection.GeneratedProtocolMessageType('KeysToSnapshotIdsEntry', (_message.Message,), {
    'DESCRIPTOR' : _LONGPOLLREQUEST_KEYSTOSNAPSHOTIDSENTRY,
    '__module__' : 'src.ray.protobuf.serve_pb2'
    # @@protoc_insertion_point(class_scope:ray.serve.LongPollRequest.KeysToSnapshotIdsEntry)
    })
  ,
  'DESCRIPTOR' : _LONGPOLLREQUEST,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.LongPollRequest)
  })
_sym_db.RegisterMessage(LongPollRequest)
_sym_db.RegisterMessage(LongPollRequest.KeysToSnapshotIdsEntry)

LongPollResult = _reflection.GeneratedProtocolMessageType('LongPollResult', (_message.Message,), {

  'UpdatedObjectsEntry' : _reflection.GeneratedProtocolMessageType('UpdatedObjectsEntry', (_message.Message,), {
    'DESCRIPTOR' : _LONGPOLLRESULT_UPDATEDOBJECTSENTRY,
    '__module__' : 'src.ray.protobuf.serve_pb2'
    # @@protoc_insertion_point(class_scope:ray.serve.LongPollResult.UpdatedObjectsEntry)
    })
  ,
  'DESCRIPTOR' : _LONGPOLLRESULT,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.LongPollResult)
  })
_sym_db.RegisterMessage(LongPollResult)
_sym_db.RegisterMessage(LongPollResult.UpdatedObjectsEntry)

EndpointInfo = _reflection.GeneratedProtocolMessageType('EndpointInfo', (_message.Message,), {

  'ConfigEntry' : _reflection.GeneratedProtocolMessageType('ConfigEntry', (_message.Message,), {
    'DESCRIPTOR' : _ENDPOINTINFO_CONFIGENTRY,
    '__module__' : 'src.ray.protobuf.serve_pb2'
    # @@protoc_insertion_point(class_scope:ray.serve.EndpointInfo.ConfigEntry)
    })
  ,
  'DESCRIPTOR' : _ENDPOINTINFO,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.EndpointInfo)
  })
_sym_db.RegisterMessage(EndpointInfo)
_sym_db.RegisterMessage(EndpointInfo.ConfigEntry)

EndpointSet = _reflection.GeneratedProtocolMessageType('EndpointSet', (_message.Message,), {

  'EndpointsEntry' : _reflection.GeneratedProtocolMessageType('EndpointsEntry', (_message.Message,), {
    'DESCRIPTOR' : _ENDPOINTSET_ENDPOINTSENTRY,
    '__module__' : 'src.ray.protobuf.serve_pb2'
    # @@protoc_insertion_point(class_scope:ray.serve.EndpointSet.EndpointsEntry)
    })
  ,
  'DESCRIPTOR' : _ENDPOINTSET,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.EndpointSet)
  })
_sym_db.RegisterMessage(EndpointSet)
_sym_db.RegisterMessage(EndpointSet.EndpointsEntry)

ActorNameList = _reflection.GeneratedProtocolMessageType('ActorNameList', (_message.Message,), {
  'DESCRIPTOR' : _ACTORNAMELIST,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.ActorNameList)
  })
_sym_db.RegisterMessage(ActorNameList)

DeploymentVersion = _reflection.GeneratedProtocolMessageType('DeploymentVersion', (_message.Message,), {
  'DESCRIPTOR' : _DEPLOYMENTVERSION,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.DeploymentVersion)
  })
_sym_db.RegisterMessage(DeploymentVersion)

ReplicaConfig = _reflection.GeneratedProtocolMessageType('ReplicaConfig', (_message.Message,), {
  'DESCRIPTOR' : _REPLICACONFIG,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.ReplicaConfig)
  })
_sym_db.RegisterMessage(ReplicaConfig)

DeploymentInfo = _reflection.GeneratedProtocolMessageType('DeploymentInfo', (_message.Message,), {
  'DESCRIPTOR' : _DEPLOYMENTINFO,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.DeploymentInfo)
  })
_sym_db.RegisterMessage(DeploymentInfo)

DeploymentRoute = _reflection.GeneratedProtocolMessageType('DeploymentRoute', (_message.Message,), {
  'DESCRIPTOR' : _DEPLOYMENTROUTE,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.DeploymentRoute)
  })
_sym_db.RegisterMessage(DeploymentRoute)

DeploymentRouteList = _reflection.GeneratedProtocolMessageType('DeploymentRouteList', (_message.Message,), {
  'DESCRIPTOR' : _DEPLOYMENTROUTELIST,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.DeploymentRouteList)
  })
_sym_db.RegisterMessage(DeploymentRouteList)

DeploymentStatusInfo = _reflection.GeneratedProtocolMessageType('DeploymentStatusInfo', (_message.Message,), {
  'DESCRIPTOR' : _DEPLOYMENTSTATUSINFO,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.DeploymentStatusInfo)
  })
_sym_db.RegisterMessage(DeploymentStatusInfo)

DeploymentStatusInfoList = _reflection.GeneratedProtocolMessageType('DeploymentStatusInfoList', (_message.Message,), {
  'DESCRIPTOR' : _DEPLOYMENTSTATUSINFOLIST,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.DeploymentStatusInfoList)
  })
_sym_db.RegisterMessage(DeploymentStatusInfoList)

ApplicationStatusInfo = _reflection.GeneratedProtocolMessageType('ApplicationStatusInfo', (_message.Message,), {
  'DESCRIPTOR' : _APPLICATIONSTATUSINFO,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.ApplicationStatusInfo)
  })
_sym_db.RegisterMessage(ApplicationStatusInfo)

StatusOverview = _reflection.GeneratedProtocolMessageType('StatusOverview', (_message.Message,), {
  'DESCRIPTOR' : _STATUSOVERVIEW,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.StatusOverview)
  })
_sym_db.RegisterMessage(StatusOverview)

PredictRequest = _reflection.GeneratedProtocolMessageType('PredictRequest', (_message.Message,), {

  'InputEntry' : _reflection.GeneratedProtocolMessageType('InputEntry', (_message.Message,), {
    'DESCRIPTOR' : _PREDICTREQUEST_INPUTENTRY,
    '__module__' : 'src.ray.protobuf.serve_pb2'
    # @@protoc_insertion_point(class_scope:ray.serve.PredictRequest.InputEntry)
    })
  ,
  'DESCRIPTOR' : _PREDICTREQUEST,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.PredictRequest)
  })
_sym_db.RegisterMessage(PredictRequest)
_sym_db.RegisterMessage(PredictRequest.InputEntry)

PredictResponse = _reflection.GeneratedProtocolMessageType('PredictResponse', (_message.Message,), {
  'DESCRIPTOR' : _PREDICTRESPONSE,
  '__module__' : 'src.ray.protobuf.serve_pb2'
  # @@protoc_insertion_point(class_scope:ray.serve.PredictResponse)
  })
_sym_db.RegisterMessage(PredictResponse)

_PREDICTAPISSERVICE = DESCRIPTOR.services_by_name['PredictAPIsService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026io.ray.serve.generatedB\013ServeProtosP\001\370\001\001'
  _REQUESTMETADATA_CONTEXTENTRY._options = None
  _REQUESTMETADATA_CONTEXTENTRY._serialized_options = b'8\001'
  _LONGPOLLREQUEST_KEYSTOSNAPSHOTIDSENTRY._options = None
  _LONGPOLLREQUEST_KEYSTOSNAPSHOTIDSENTRY._serialized_options = b'8\001'
  _LONGPOLLRESULT_UPDATEDOBJECTSENTRY._options = None
  _LONGPOLLRESULT_UPDATEDOBJECTSENTRY._serialized_options = b'8\001'
  _ENDPOINTINFO_CONFIGENTRY._options = None
  _ENDPOINTINFO_CONFIGENTRY._serialized_options = b'8\001'
  _ENDPOINTSET_ENDPOINTSENTRY._options = None
  _ENDPOINTSET_ENDPOINTSENTRY._serialized_options = b'8\001'
  _PREDICTREQUEST_INPUTENTRY._options = None
  _PREDICTREQUEST_INPUTENTRY._serialized_options = b'8\001'
  _DEPLOYMENTLANGUAGE._serialized_start=4036
  _DEPLOYMENTLANGUAGE._serialized_end=4078
  _DEPLOYMENTSTATUS._serialized_start=4080
  _DEPLOYMENTSTATUS._serialized_end=4194
  _APPLICATIONSTATUS._serialized_start=4197
  _APPLICATIONSTATUS._serialized_end=4423
  _AUTOSCALINGCONFIG._serialized_start=44
  _AUTOSCALINGCONFIG._serialized_end=505
  _DEPLOYMENTCONFIG._serialized_start=508
  _DEPLOYMENTCONFIG._serialized_end=1173
  _REQUESTMETADATA._serialized_start=1176
  _REQUESTMETADATA._serialized_end=1412
  _REQUESTMETADATA_CONTEXTENTRY._serialized_start=1354
  _REQUESTMETADATA_CONTEXTENTRY._serialized_end=1412
  _REQUESTWRAPPER._serialized_start=1414
  _REQUESTWRAPPER._serialized_end=1450
  _UPDATEDOBJECT._serialized_start=1452
  _UPDATEDOBJECT._serialized_end=1541
  _LONGPOLLREQUEST._serialized_start=1544
  _LONGPOLLREQUEST._serialized_end=1731
  _LONGPOLLREQUEST_KEYSTOSNAPSHOTIDSENTRY._serialized_start=1663
  _LONGPOLLREQUEST_KEYSTOSNAPSHOTIDSENTRY._serialized_end=1731
  _LONGPOLLRESULT._serialized_start=1734
  _LONGPOLLRESULT._serialized_end=1931
  _LONGPOLLRESULT_UPDATEDOBJECTSENTRY._serialized_start=1840
  _LONGPOLLRESULT_UPDATEDOBJECTSENTRY._serialized_end=1931
  _ENDPOINTINFO._serialized_start=1934
  _ENDPOINTINFO._serialized_end=2127
  _ENDPOINTINFO_CONFIGENTRY._serialized_start=2070
  _ENDPOINTINFO_CONFIGENTRY._serialized_end=2127
  _ENDPOINTSET._serialized_start=2130
  _ENDPOINTSET._serialized_end=2299
  _ENDPOINTSET_ENDPOINTSENTRY._serialized_start=2214
  _ENDPOINTSET_ENDPOINTSENTRY._serialized_end=2299
  _ACTORNAMELIST._serialized_start=2301
  _ACTORNAMELIST._serialized_end=2338
  _DEPLOYMENTVERSION._serialized_start=2341
  _DEPLOYMENTVERSION._serialized_end=2513
  _REPLICACONFIG._serialized_start=2516
  _REPLICACONFIG._serialized_end=2724
  _DEPLOYMENTINFO._serialized_start=2727
  _DEPLOYMENTINFO._serialized_end=3054
  _DEPLOYMENTROUTE._serialized_start=3056
  _DEPLOYMENTROUTE._serialized_end=3163
  _DEPLOYMENTROUTELIST._serialized_start=3165
  _DEPLOYMENTROUTELIST._serialized_end=3259
  _DEPLOYMENTSTATUSINFO._serialized_start=3261
  _DEPLOYMENTSTATUSINFO._serialized_end=3382
  _DEPLOYMENTSTATUSINFOLIST._serialized_start=3384
  _DEPLOYMENTSTATUSINFOLIST._serialized_end=3499
  _APPLICATIONSTATUSINFO._serialized_start=3502
  _APPLICATIONSTATUSINFO._serialized_end=3656
  _STATUSOVERVIEW._serialized_start=3659
  _STATUSOVERVIEW._serialized_end=3846
  _PREDICTREQUEST._serialized_start=3849
  _PREDICTREQUEST._serialized_end=3983
  _PREDICTREQUEST_INPUTENTRY._serialized_start=3927
  _PREDICTREQUEST_INPUTENTRY._serialized_end=3983
  _PREDICTRESPONSE._serialized_start=3985
  _PREDICTRESPONSE._serialized_end=4034
  _PREDICTAPISSERVICE._serialized_start=4425
  _PREDICTAPISSERVICE._serialized_end=4511
# @@protoc_insertion_point(module_scope)
