# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/job_agent.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import agent_manager_pb2 as src_dot_ray_dot_protobuf_dot_agent__manager__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n src/ray/protobuf/job_agent.proto\x12\x07ray.rpc\x1a$src/ray/protobuf/agent_manager.proto\"B\n\x17InitializeJobEnvRequest\x12\'\n\x0fjob_description\x18\x01 \x01(\tR\x0ejobDescription\"\x8c\x01\n\x15InitializeJobEnvReply\x12/\n\x06status\x18\x01 \x01(\x0e\x32\x17.ray.rpc.AgentRpcStatusR\x06status\x12#\n\rerror_message\x18\x02 \x01(\tR\x0c\x65rrorMessage\x12\x1d\n\ndriver_pid\x18\x03 \x01(\x05R\tdriverPid2g\n\x0fJobAgentService\x12T\n\x10InitializeJobEnv\x12 .ray.rpc.InitializeJobEnvRequest\x1a\x1e.ray.rpc.InitializeJobEnvReplyB\x03\xf8\x01\x01\x62\x06proto3')



_INITIALIZEJOBENVREQUEST = DESCRIPTOR.message_types_by_name['InitializeJobEnvRequest']
_INITIALIZEJOBENVREPLY = DESCRIPTOR.message_types_by_name['InitializeJobEnvReply']
InitializeJobEnvRequest = _reflection.GeneratedProtocolMessageType('InitializeJobEnvRequest', (_message.Message,), {
  'DESCRIPTOR' : _INITIALIZEJOBENVREQUEST,
  '__module__' : 'src.ray.protobuf.job_agent_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.InitializeJobEnvRequest)
  })
_sym_db.RegisterMessage(InitializeJobEnvRequest)

InitializeJobEnvReply = _reflection.GeneratedProtocolMessageType('InitializeJobEnvReply', (_message.Message,), {
  'DESCRIPTOR' : _INITIALIZEJOBENVREPLY,
  '__module__' : 'src.ray.protobuf.job_agent_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.InitializeJobEnvReply)
  })
_sym_db.RegisterMessage(InitializeJobEnvReply)

_JOBAGENTSERVICE = DESCRIPTOR.services_by_name['JobAgentService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _INITIALIZEJOBENVREQUEST._serialized_start=83
  _INITIALIZEJOBENVREQUEST._serialized_end=149
  _INITIALIZEJOBENVREPLY._serialized_start=152
  _INITIALIZEJOBENVREPLY._serialized_end=292
  _JOBAGENTSERVICE._serialized_start=294
  _JOBAGENTSERVICE._serialized_end=397
# @@protoc_insertion_point(module_scope)
