<div class='titleData jp-RenderedHTMLCommon'>
  <h3>{{ title }}</h3>
  {{ data }}
</div>
<style>
.titleData h3 {
    border-bottom-width: var(--jp-border-width);
    border-bottom-color: var(--jp-border-color0);
    border-bottom-style: solid;
}
</style>
